import logging
import typing as t
from pathlib import Path

from yarl import URL

from influxio.adapter import FileAdapter, InfluxDbApiAdapter, InfluxDbEngineAdapter, SqlAlchemyAdapter
from influxio.model import CommandResult
from influxio.util.common import url_fullpath
from influxio.util.db import get_sqlalchemy_dialects

logger = logging.getLogger(__name__)


def copy(source: str, target: str, progress: bool = False) -> t.Union[CommandResult, None]:
    """
    Copy/transfer data from/to InfluxDB API / InfluxDB line protocol / RDBMS.

    RDBMS is any SQLAlchemy-supported database.

    `source` and `target` are resource identifiers in URL format.

    When InfluxDB is addressed, the schema is:
    http://example:token@localhost:8086/testdrive/demo

    This means:
    - Organization: example
    - Authentication: token
    - Bucket: testdrive
    - Measurement: demo

    When an RDBMS is addressed through SQLAlchemy, the schema is:
    http://username:password@localhost:12345/testdrive/demo

    This means:
    - Database or schema: testdrive
    - Table name: demo
    """
    source_url = URL(source)
    target_url = URL(target)

    sqlalchemy_dialects = get_sqlalchemy_dialects()

    logger.info(f"Copying from {source} to {target}")

    scheme_primary = target_url.scheme.split("+")[0]

    if target_url.scheme.startswith("http"):
        sink = InfluxDbApiAdapter.from_url(target)
    elif scheme_primary in sqlalchemy_dialects:
        sink = SqlAlchemyAdapter.from_url(target, progress=True)
    elif target_url.scheme == "file":
        sink = FileAdapter.from_url(target, progress=True)
    else:
        raise NotImplementedError(f"Data sink not implemented: {target_url}")

    if source_url.scheme == "testdata":
        from influxio.testdata import DataFrameFactory

        dff = DataFrameFactory(**source_url.query)
        df = dff.make(source_url.host)
        sink.write_df(df)

    elif source_url.scheme == "file":

        # Export
        if target_url.scheme == "file":
            path = url_fullpath(source_url)
            source_path_dir = [path.name for path in Path(path).iterdir()]
            if "data" in source_path_dir and "wal" in source_path_dir:
                source_element = InfluxDbEngineAdapter.from_url(source)
                if not source_element.bucket_id:
                    raise ValueError("Parameter missing or empty: bucket-id")
                if not source_element.measurement:
                    raise ValueError("Parameter missing or empty: measurement")
                return source_element.to_lineprotocol(url=target_url)
            else:
                raise FileNotFoundError(f"No InfluxDB data directory: {path}")

        # Import
        else:
            path = Path(source_url.host).joinpath(Path(source_url.path).relative_to("/"))
            # TODO: Determine file type by suffix.
            # TODO: Make `precision` configurable.
            sink.from_lineprotocol(path)

    elif source_url.scheme.startswith("http"):
        # TODO: Improve dispatching.
        source_url_str = str(source_url)
        if isinstance(sink, (FileAdapter, SqlAlchemyAdapter)) and ".lp" not in source_url_str:
            source_node = InfluxDbApiAdapter.from_url(source)
            sink.write(source_node)
        else:
            sink.from_lineprotocol(source_url_str)

    else:
        raise NotImplementedError(f"Data source not implemented: {source_url}")

    return None



import io
import json
import logging
import sys
import typing as t
from pathlib import Path

import influxdb_client.rest
import pandas as pd
import polars as pl
import psycopg2
import sqlalchemy
import sqlalchemy as sa
from fsspec import filesystem
from influxdb_client import InfluxDBClient
from sqlalchemy_utils import create_database
from upath import UPath
from yarl import URL

from influxio.io import dataframe_to_lineprotocol, dataframe_to_sql, dataframes_from_lineprotocol
from influxio.model import CommandResult, DataFormat, OutputFile
from influxio.util.common import run_command, url_fullpath

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 60.0


class InfluxDbApiAdapter:
    def __init__(
        self,
        url: str,
        token: str,
        org: str,
        bucket: str,
        measurement: str,
        debug: bool = False,
        timeout: float = DEFAULT_TIMEOUT,
    ):
        self.url = url
        self.token = token
        self.org = org
        self.bucket = bucket
        self.measurement = measurement
        self.debug = debug
        self.timeout = timeout
        self.client = InfluxDBClient(
            url=self.url, org=self.org, token=self.token, debug=self.debug, timeout=int(self.timeout * 1000.0)
        )

    @classmethod
    def from_url(cls, url: t.Union[URL, str], **kwargs) -> "InfluxDbApiAdapter":
        if isinstance(url, str):
            url: URL = URL(url)
        token = url.password
        org = url.user
        try:
            bucket, measurement = url.path.strip("/").split("/")
        except ValueError:
            bucket = url.path.strip("/").split("/")[0]
            measurement = None
        bare_url = f"{url.scheme}://{url.host}:{url.port}"
        kwargs.setdefault("timeout", float(url.query.get("timeout", DEFAULT_TIMEOUT)))
        return cls(url=bare_url, token=token, org=org, bucket=bucket, measurement=measurement, **kwargs)

    def delete_measurement(self):
        """
        https://docs.influxdata.com/influxdb/v2/write-data/delete-data/
        """
        try:
            return self.client.delete_api().delete(
                start="1677-09-21T00:12:43.145224194Z",
                stop="2262-04-11T23:47:16.854775806Z",
                predicate=f'_measurement="{self.measurement}"',
                bucket=self.bucket,
            )
        except influxdb_client.rest.ApiException as ex:
            if ex.status != 404:
                raise

    def read_df(self):
        """ """
        filter_expression = ""
        if self.measurement is not None:
            filter_expression = f'|> filter(fn: (r) => r._measurement == "{self.measurement}")'
        query = f"""
        from(bucket:"{self.bucket}")
            |> range(start: 0, stop: now())
            {filter_expression}
            |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value")
        """
        for df in self.client.query_api().query_data_frame_stream(query=query):
            df = df.drop(["result", "table", "_start", "_stop"], axis=1)
            df = df.rename(columns={"_time": "time", "_measurement": "measurement"})
            yield df

    def read_records(self, bucket: t.Optional[str] = None, measurement: t.Optional[str] = None) -> t.Dict[str, t.Any]:
        bucket = bucket or self.bucket
        measurement = measurement or self.measurement
        query = f"""
            from(bucket: "{bucket}")
                |> range(start: 0)
                |> filter(fn: (r) => r._measurement == "{measurement}")
            """
        result = self.client.query_api().query(query=query)
        return json.loads(result.to_json())

    def ensure_bucket(self):
        try:
            self.client.buckets_api().create_bucket(bucket_name=self.bucket)
        except influxdb_client.rest.ApiException as ex:
            if ex.status == 422:
                pass
            else:
                raise
        logger.info(f"Bucket id is {self.get_bucket_id()}")

    def delete_bucket(self, missing_ok: bool = True):
        """
        https://docs.influxdata.com/influxdb/v2/admin/buckets/delete-bucket/
        """
        try:
            bucket_id = self.get_bucket_id()
        except KeyError:
            if missing_ok:
                return
            else:
                raise
        try:
            self.client.buckets_api().delete_bucket(bucket_id)
        except influxdb_client.rest.ApiException as ex:
            if ex.status == 404 and missing_ok:
                pass
            else:
                raise

    def write_df(self, df: pd.DataFrame):
        """
        Use batching API to import data frame into InfluxDB.

        https://github.com/influxdata/influxdb-client-python/blob/master/examples/ingest_large_dataframe.py

        TODO: Add precision.
        """
        logger.info(f"Importing data frame to InfluxDB. bucket={self.bucket}, measurement={self.measurement}")
        self.ensure_bucket()
        with self.client.write_api() as write_api:
            write_api.write(
                bucket=self.bucket,
                record=df,
                data_frame_measurement_name=self.measurement,
                # TODO: Add more parameters.
                # write_precision=WritePrecision.MS,  # noqa: ERA001
                # data_frame_tag_columns=['tag'],  # noqa: ERA001
            )

    def from_lineprotocol(self, source: t.Union[Path, str], precision: str = "ns"):
        """
        Import data from file in lineprotocol format (ILP) by invoking `influx write`.

        Precision of the timestamps of the lines (default: ns) [$INFLUX_PRECISION]

        The default precision for timestamps is in nanoseconds. If the precision of
        the timestamps is anything other than nanoseconds (ns), you must specify the
        precision in your write request. InfluxDB accepts the following precisions:

            ns - Nanoseconds
            us - Microseconds
            ms - Milliseconds
            s - Seconds

        -- https://docs.influxdata.com/influxdb/v2/reference/syntax/line-protocol/
        """
        is_url = False
        try:
            URL(source)
            is_url = True
        except Exception:  # noqa: S110
            pass

        logger.info(f"Importing line protocol format to InfluxDB. bucket={self.bucket}")
        self.ensure_bucket()

        if is_url:
            source_option = f'--url="{str(source)}"'
        else:
            source_option = f'--file="{str(source)}"'
        command = f"""
        influx write \
            --host="{self.url}" \
            --token="{self.token}" \
            --org="{self.org}" \
            --bucket="{self.bucket}" \
            --precision={precision} \
            --format=lp \
            {source_option}
        """
        # print("command:", command)  # noqa: ERA001
        run_command(command)

    @property
    def bucket_id(self) -> str:
        return self.get_bucket_id()

    def get_bucket_id(self) -> str:
        """
        Resolve bucket name to bucket id.
        """
        bucket: influxdb_client.Bucket = self.client.buckets_api().find_bucket_by_name(bucket_name=self.bucket)
        if bucket is None:
            raise KeyError(f"Bucket not found: {self.bucket}")
        return bucket.id


class InfluxDbEngineAdapter:
    def __init__(self, path: t.Union[Path, str], bucket_id: str, measurement: str, debug: bool = False):

        if isinstance(path, str):
            path: Path = Path(path)

        self.path = path
        self.bucket_id = bucket_id
        self.measurement = measurement
        self.debug = debug

    @classmethod
    def from_url(cls, url: t.Union[URL, str], **kwargs) -> "InfluxDbEngineAdapter":
        if isinstance(url, str):
            url: URL = URL(url)
        return cls(
            path=url_fullpath(url),
            bucket_id=url.query.get("bucket-id"),
            measurement=url.query.get("measurement"),
            **kwargs,
        )

    def to_lineprotocol(self, url: t.Union[URL, str]) -> CommandResult:
        """
        Export data into lineprotocol format (ILP) by invoking `influxd inspect export-lp`.

        TODO: Unify with `FileAdapter` sink and expand with `InfluxDbApiAdapter`'s API connectivity.
        TODO: Using a hyphen `-` for `--output-path` works well now, so export can also go to stdout.
        TODO: By default, it will *append* to the .lp file.
              Make it configurable to "replace" data.
        TODO: Make it configurable to use compression, or not.
        TODO: Propagate `--start` and `--end` parameters.
        TODO: Capture stderr messages, and forward user admonition.
              »detected deletes in WAL file, some deleted data may be brought back by replaying this export«
              -- https://github.com/influxdata/influxdb/issues/24456

        https://docs.influxdata.com/influxdb/v2.6/migrate-data/migrate-oss/
        """
        if isinstance(url, str):
            url: URL = URL(url)
        format_ = DataFormat.from_url(url)
        logger.info(f"Exporting data to InfluxDB line protocol format (ILP): {format_}")
        command = f"""
        influxd inspect export-lp \
            --engine-path '{self.path}' \
            --bucket-id '{self.bucket_id}' \
            --measurement '{self.measurement}' \
            --output-path '{url_fullpath(url)}'
        """.rstrip()
        if format_ is DataFormat.LINE_PROTOCOL_COMPRESSED:
            command += " --compress"
        out = run_command(command)

        stderr = out.stderr.decode("utf-8")

        # Decode output of `influxd inspect export-lp`.
        """
        {"level":"info","ts":1712536769.359062,"caller":"export_lp/export_lp.go:219","msg":"exporting TSM files","tsm_dir":"var/lib/influxdb2/engine/data/372d1908eab801a6","file_count":3}
        {"level":"info","ts":1712536769.3782709,"caller":"export_lp/export_lp.go:315","msg":"exporting WAL files","wal_dir":"var/lib/influxdb2/engine/wal/372d1908eab801a6","file_count":3}
        {"level":"info","ts":1712536769.3783438,"caller":"export_lp/export_lp.go:204","msg":"export complete"}
        """  # noqa: E501
        if format_ in [DataFormat.LINE_PROTOCOL_UNCOMPRESSED, DataFormat.LINE_PROTOCOL_COMPRESSED]:
            report = pd.read_json(path_or_buf=io.StringIO(stderr), lines=True).to_dict(orient="records")
            tsm_file_count = report[0]["file_count"]
            wal_file_count = report[1]["file_count"]
            if tsm_file_count == 0 and wal_file_count == 0:
                raise FileNotFoundError(r"Export yielded zero records. Make sure to use a valid bucket-id.")
        else:
            raise NotImplementedError(f"Format is not supported: {format_}")
        if out.stdout:
            sys.stdout.buffer.write(out.stdout)
        return CommandResult(stderr=stderr, exitcode=out.returncode)


class SqlAlchemyAdapter:
    """
    Adapter to talk to SQLAlchemy-compatible databases.
    """

    def __init__(self, url: t.Union[URL, str], progress: bool = False, debug: bool = False):
        self.progress = progress

        if isinstance(url, str):
            url: URL = URL(url)

        self.database, self.table = self.decode_database_table(url)
        if self.database:
            self.table_fqn = f"{self.database}.{self.table}"
            self.table_fqn_quoted = f'"{self.database}"."{self.table}"'
        else:
            self.table_fqn = self.table
            self.table_fqn_quoted = f'"{self.table}"'
        self.if_exists = url.query.get("if-exists") or "fail"

        # Special handling for SQLite and CrateDB databases.
        self.dburi = str(url.with_query(None))
        if url.scheme == "crate":
            query_args_passthrough = ["ssl"]
            query = url.query
            url = url.with_path("")
            if self.database:
                url = url.update_query({"schema": self.database})
                for arg in query_args_passthrough:
                    if arg in query:
                        url = url.update_query({arg: query[arg]})
            self.dburi = str(url)
        elif url.scheme == "sqlite":
            self.dburi = self.dburi.replace("sqlite:/", "sqlite:///")
        else:
            url = url.with_path(self.database)
            self.dburi = str(url)

        logger.info(f"SQLAlchemy DB URI: {self.dburi}")

    @classmethod
    def from_url(cls, url: t.Union[URL, str], **kwargs) -> "SqlAlchemyAdapter":
        return cls(url=url, **kwargs)

    def write(self, source: t.Union[pd.DataFrame, InfluxDbApiAdapter], table: t.Optional[str] = None):
        table = table or self.table
        table_fqn = f"{self.database}.{table}"
        table_fqn_quoted = f'"{self.database}"."{table}"'
        logger.info("Loading dataframes into RDBMS/SQL database using pandas/Dask")
        logger.info(f"Target table: {table_fqn_quoted}")

        # For CrateDB, converge `if-exists={fail,replace}` to `if-exists=append`, and create the table manually.
        if self.dburi.startswith("crate") and self.if_exists in ["fail", "replace"]:

            # Prevent overwriting existing table when `if-exists=fail`.
            if self.if_exists == "fail" and self.table_exists():
                raise ValueError(f"Table '{table_fqn}' already exists.")

            # Create table with dynamic column policy, to accompany new tags that
            # progressively drop in while processing the whole dataset.
            engine = sa.create_engine(self.dburi, echo=True)
            with engine.connect() as connection:
                connection.execute(sa.text(f"DROP TABLE IF EXISTS {table_fqn_quoted}"))
                connection.execute(
                    sa.text(f"CREATE TABLE {table_fqn_quoted} (dummy int) WITH (column_policy='dynamic')")
                )

            # Because the table has been created already, switch to `append` mode.
            self.if_exists = "append"

        if isinstance(source, InfluxDbApiAdapter):
            logger.info("Loading data from InfluxDB API")
            has_data = False
            for df in source.read_df():
                has_data = True
                dataframe_to_sql(
                    df,
                    dburi=self.dburi,
                    tablename=table,
                    schema=self.database,
                    if_exists=self.if_exists,
                    progress=self.progress,
                )
            if not has_data:
                msg = "No data has been loaded from InfluxDB"
                logger.error(msg)
                raise IOError(msg)
        elif isinstance(source, (pd.DataFrame, pl.DataFrame)):
            logger.info("Loading data from dataframe")
            dataframe_to_sql(
                source,
                dburi=self.dburi,
                tablename=table,
                schema=self.database,
                if_exists=self.if_exists,
                progress=self.progress,
            )
        else:
            raise NotImplementedError(f"Failed handling source: {source}")

    def refresh_table(self):
        engine = sa.create_engine(self.dburi)
        with engine.connect() as connection:
            return connection.execute(sa.text(f"REFRESH TABLE {self.table_fqn_quoted};"))

    def table_exists(self):
        engine = sa.create_engine(self.dburi)
        metadata = sa.MetaData()
        metadata.reflect(bind=engine, schema=self.database)
        return self.table_fqn in metadata.tables

    def read_records(self, table: t.Optional[str] = None) -> t.List[t.Dict]:
        table = table or self.table_fqn_quoted
        engine = sa.create_engine(self.dburi)
        with engine.connect() as connection:
            result = connection.execute(sa.text(f"SELECT * FROM {table};"))  # noqa: S608
            records = [dict(item) for item in result.mappings().fetchall()]
            return records

    def create_database(self):
        try:
            return create_database(self.dburi)
        except sqlalchemy.exc.ProgrammingError as ex:
            if "psycopg2.errors.DuplicateDatabase" not in str(ex):
                raise

    def run_sql(self, sql: str):
        engine = sa.create_engine(self.dburi)
        with engine.connect() as connection:
            if hasattr(connection.connection, "set_isolation_level"):
                connection.connection.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
            return connection.execute(sa.text(sql))

    def run_sql_raw(self, sql: str):
        engine = sa.create_engine(self.dburi)
        connection = engine.raw_connection()
        if hasattr(connection, "set_isolation_level"):
            connection.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        cursor = connection.cursor()
        cursor.execute(sql)
        result = cursor.fetchall()
        cursor.close()
        connection.close()
        return result

    @staticmethod
    def decode_database_table(url: URL):
        """
        Decode database and table names from database URI path and/or query string.

        Variants:

            /<database>
            /<database>/<table>
            ?database=<database>&table=<table>
            ?schema=<database>&table=<table>
        """
        database = None
        table = None
        path = url.path.strip("/")
        try:
            database, table = path.split("/")
        except ValueError as ex:
            if url.host and "not enough values to unpack" in str(ex):
                database = path or None
            if "too many values to unpack" not in str(ex) and "not enough values to unpack" not in str(ex):
                raise
            database = url.query.get("database", database)
            table = url.query.get("table", table)
            if url.scheme == "crate" and not database:
                database = url.query.get("schema", database)
        return database, table

    def from_lineprotocol(self, source: t.Union[Path, str], precision: str = "ns"):
        """
        Load data from file or resource in lineprotocol format (ILP).
        """
        logger.info(f"Loading line protocol data. source={source}")
        p = UPath(source)
        fs = filesystem(p.protocol, **p.storage_options)  # equivalent to p.fs
        with fs.open(p.path) as fp:
            frames = dataframes_from_lineprotocol(fp)
            for table, df in frames.items():
                self.write(df, table=table)


class FileAdapter:
    """
    Adapter for pipelining data in and out of files.
    """

    def __init__(self, url: t.Union[URL, str], progress: bool = False, debug: bool = False):
        self.progress = progress

        if isinstance(url, str):
            url: URL = URL(url)

        self.output = OutputFile.from_url(url)

    @classmethod
    def from_url(cls, url: t.Union[URL, str], **kwargs) -> "FileAdapter":
        """
        Factory to create a `FileAdapter` instance from a URL.
        """
        return cls(url=url, **kwargs)

    def write(self, source: t.Union[pd.DataFrame, InfluxDbApiAdapter]):
        """
        Export data from a pandas DataFrame or from an API-connected InfluxDB database into lineprotocol format (ILP).
        """
        logger.info(f"Exporting dataframes in {self.output.format.value} format to {self.output.path}")
        generators = []
        if self.output.format is DataFormat.LINE_PROTOCOL_UNCOMPRESSED:
            if isinstance(source, InfluxDbApiAdapter):
                for df in source.read_df():
                    generators.append(dataframe_to_lineprotocol(df, progress=self.progress))
            elif isinstance(source, pd.DataFrame):
                generators.append(dataframe_to_lineprotocol(source, progress=self.progress))
            else:
                raise NotImplementedError(f"Unknown data source: {source}")
        else:
            raise NotImplementedError(f"File output format not implemented: {self.output.format}")
        for generator in generators:
            print("\n".join(generator), file=self.output.stream)



import platform
import textwrap
import typing as t

from colorama import Fore, Style

string_or_list = t.Union[str, t.List[str]]


def section(text: str):
    return armor(text, title)


def subsection(text: str):
    return armor(text, subtitle)


def armor(text: str, formatter: t.Callable):
    guard = "=" * len(text)
    print(guard)
    print(formatter(text))
    print(guard)


def title(text: str):
    return Fore.CYAN + Style.BRIGHT + text + Style.RESET_ALL


def subtitle(text: str):
    return Fore.YELLOW + Style.BRIGHT + text + Style.RESET_ALL


def text_list(data: string_or_list):
    if isinstance(data, str):
        return data
    elif isinstance(data, t.List):
        return ", ".join(data)
    else:
        raise ValueError("Unable to flatten list")


def bullet_list(data: t.List[str]):
    data.sort()
    lines = [f"- {text_list(line)}" for line in data]
    return "\n".join(lines)


def wrap_list(data: t.List[str], **kwargs):
    data = [item for item in data if item is not None]
    data.sort()
    text = ", ".join(data)
    return "\n".join(textwrap.wrap(text, width=80, **kwargs))


def bullet_item(data: string_or_list, label: str = None):
    if data is None:
        return None
    if label is None:
        label = ""
    else:
        label = f"{label}: "
    if isinstance(data, (t.List, t.KeysView, t.ValuesView)):
        text = wrap_list(list(data), subsequent_indent="  ")
    else:
        text = data
    return f"- {label}{text}"


class AboutReport:
    @staticmethod
    def platform():
        section("Platform")
        print()

        subsection("Python")
        print(bullet_item(platform.platform()))
        print()

        # SQLAlchemy
        import sqlalchemy.dialects

        from influxio.util.compat import entry_points

        subsection("SQLAlchemy")
        print(bullet_item(list(sqlalchemy.dialects.__all__), label="Dialects built-in"))
        eps = entry_points(group="sqlalchemy.dialects")
        dialects = [dialect.name for dialect in eps]
        print(bullet_item(dialects, label="Dialects 3rd-party"))
        print(bullet_item(sqlalchemy.dialects.plugins.impls.keys(), label="Plugins"))
        print()

        # fsspec
        import fsspec

        subsection("fsspec protocols")
        print(bullet_item(fsspec.available_protocols()))
        print()

        subsection("fsspec compressions")
        print(bullet_item(fsspec.available_compressions()))
        print()

        # pandas
        subsection("pandas module versions")
        import pandas

        pandas.show_versions(as_json=False)
        print()



import json
import logging
import shlex
import subprocess
import sys
import typing as t
from enum import Enum
from textwrap import dedent

from yarl import URL

logger = logging.getLogger(__name__)


def setup_logging(level=logging.INFO):
    log_format = "%(asctime)-15s [%(name)-24s] %(levelname)-8s: %(message)s"
    logging.basicConfig(format=log_format, stream=sys.stderr, level=level)


def get_version(appname):
    from importlib.metadata import PackageNotFoundError, version  # noqa

    try:
        return version(appname)
    except PackageNotFoundError:  # pragma: no cover
        return "unknown"


def jd(data: t.Any):
    """
    Pretty-print JSON with indentation.
    """
    print(json.dumps(data, indent=2))  # noqa: T201


def run_command(command: str) -> subprocess.CompletedProcess:
    """
    https://stackoverflow.com/a/48813330
    """
    command = dedent(command).strip()
    cmd = shlex.split(command)
    logger.info(f"Running command: {command}")
    try:
        output: subprocess.CompletedProcess = subprocess.run(cmd, check=True, capture_output=True)  # noqa: S603
    except subprocess.CalledProcessError as exc:
        logger.error(f"Command failed (exit code {exc.returncode}). The command was:\n{command}")
        logger.error(exc.stderr.decode("utf-8"))
        raise
    else:
        if output:
            stderr = output.stderr.decode("utf-8")
            if stderr:
                logger.info(f"Command stderr:\n{stderr}")
        return output


class AutoStrEnum(str, Enum):
    """
    StrEnum where enum.auto() returns the field name.
    See https://docs.python.org/3.9/library/enum.html#using-automatic-values

    From https://stackoverflow.com/a/74539097.
    """

    @staticmethod
    def _generate_next_value_(name: str, start: int, count: int, last_values: list) -> str:  # noqa: ARG004
        return name


def url_fullpath(url: URL):
    fullpath = url.host or ""
    if fullpath == "-":
        return fullpath
    fullpath += url.path or ""
    return fullpath.rstrip("/")


def url_or_path(url: URL):
    return url.host or url.path



import logging

import click

import influxio.core
from influxio.util.cli import boot_click, docstring_format_verbatim
from influxio.util.report import AboutReport

logger = logging.getLogger(__name__)


def help_copy():
    """
    Import and export data into/from InfluxDB

    SOURCE can be a file or a URL.
    TARGET can be a file or a URL.

    Synopsis
    ========

    # Export from API to database.
    influxio copy \
        "http://example:token@localhost:8086/testdrive/demo" \
        "sqlite://export.sqlite?table=demo"

    # Export from data directory to line protocol format.
    influxio copy \
        "file:///path/to/influxdb/engine?bucket-id=372d1908eab801a6&measurement=demo" \
        "file://export.lp"

    Examples
    ========

    Export from API
    ---------------

    # From API to database file.
    influxio copy \
        http://example:token@localhost:8086/testdrive/demo \
        sqlite:///export.sqlite

    # From API to database server.
    influxio copy \
        http://example:token@localhost:8086/testdrive/demo \
        crate://crate@localhost:4200/testdrive

    # From API to line protocol file.
    influxio copy \
        http://example:token@localhost:8086/testdrive/demo \
        file://export.lp

    # From API to line protocol on stdout.
    influxio copy \
        "http://example:token@localhost:8086/testdrive/demo" \
        "file://-?format=lp"

    Export from data directory
    --------------------------

    # From InfluxDB data directory to line protocol file.
    influxio copy \
        "file:///path/to/influxdb/engine?bucket-id=372d1908eab801a6&measurement=demo" \
        "file://export.lp"

    # From InfluxDB data directory to line protocol file, compressed with gzip.
    influxio copy \
        "file:///path/to/influxdb/engine?bucket-id=372d1908eab801a6&measurement=demo" \
        "file://export.lp.gz"

    # From InfluxDB data directory to line protocol on stdout.
    influxio copy \
        "file:///path/to/influxdb/engine?bucket-id=372d1908eab801a6&measurement=demo" \
        ""file://-?format=lp"

    Convert from file
    -----------------

    # From line protocol file to database.
    influxio copy \
        "file://export.lp" \
        "sqlite://export.sqlite?table=export"


    Import
    ------

    # From test data to API.
    # Choose one of dummy, mixed, dateindex, wide.
    influxio copy \
        "testdata://dateindex/" \
        "http://example:token@localhost:8086/testdrive/demo"

    # With selected amount of rows.
    influxio copy \
        "testdata://dateindex/?rows=42" \
        "http://example:token@localhost:8086/testdrive/demo"

    # With selected amount of rows and columns (only supported by certain test data sources).
    influxio copy \
        "testdata://wide/?rows=42&columns=42" \
        "http://example:token@localhost:8086/testdrive/demo"

    # From line protocol file to InfluxDB API.
    influxio copy \
        "file://tests/testdata/basic.lp" \
        "http://example:token@localhost:8086/testdrive/demo"

    # From line protocol file to InfluxDB API.
    influxio copy \
        "https://github.com/influxdata/influxdb2-sample-data/raw/master/air-sensor-data/air-sensor-data.lp" \
        "http://example:token@localhost:8086/testdrive/demo"

    # From line protocol file to any database supported by SQLAlchemy.
    influxio copy \
        "file://export.lp" \
        "sqlite://export.sqlite?table=export"

    Documentation
    =============

    More options and examples can be discovered on the influxio README [1].

    [1] https://github.com/daq-tools/influxio/blob/main/README.rst
    """  # noqa: E501


@click.group()
@click.version_option(package_name="influxio")
@click.option("--verbose", is_flag=True, required=False, default=True, help="Turn on logging")
@click.option("--debug", is_flag=True, required=False, help="Turn on logging with debug level")
@click.pass_context
def cli(ctx: click.Context, verbose: bool, debug: bool):
    return boot_click(ctx, verbose, debug)


@cli.command("info", help="Report about platform information")
def info():
    AboutReport.platform()


@cli.command(
    "copy",
    help=docstring_format_verbatim(help_copy.__doc__),
    context_settings={"max_content_width": 120},
)
@click.argument("source", type=str, required=True)
@click.argument("target", type=str, required=True)
@click.pass_context
def copy(ctx: click.Context, source: str, target: str):
    influxio.core.copy(source, target)
    logger.info("Ready.")



def get_sqlalchemy_dialects():
    """
    Return list of available SQLAlchemy dialects.

    TODO: Synchronize with influxio.util.report.
    """
    import sqlalchemy.dialects

    from influxio.util.compat import entry_points

    dialects = list(sqlalchemy.dialects.__all__)
    eps = entry_points(group="sqlalchemy.dialects")
    dialects += [dialect.name for dialect in eps]
    return sorted(set(dialects))



from contextlib import contextmanager
from unittest.mock import patch

from pandas.io.sql import SQLTable

_create_table_setup_dist = SQLTable._create_table_setup


@contextmanager
def sqlalchemy_table_kwargs(**kwargs):
    def _create_table_setup(self):
        table = _create_table_setup_dist(self)
        table.kwargs.update(kwargs)
        return table

    with patch("pandas.io.sql.SQLTable._create_table_setup", _create_table_setup):
        yield



import dataclasses
import sys
import typing as t
from enum import auto

from yarl import URL

from influxio.util.common import AutoStrEnum, url_fullpath, url_or_path


@dataclasses.dataclass
class CommandResult:
    exitcode: int
    stderr: str


class DataFormat(AutoStrEnum):
    """
    Manage data formats implemented by `FileAdapter`.
    """

    LINE_PROTOCOL_UNCOMPRESSED = auto()
    LINE_PROTOCOL_COMPRESSED = auto()
    ANNOTATED_CSV = auto()

    @classmethod
    def from_name(cls, path: str) -> "DataFormat":
        path = path.rstrip("/")
        if path.endswith(".lp"):
            return cls.LINE_PROTOCOL_UNCOMPRESSED
        elif path.endswith(".lp.gz"):
            return cls.LINE_PROTOCOL_COMPRESSED
        elif path.endswith(".csv"):
            return cls.ANNOTATED_CSV
        else:
            raise ValueError(f"Unable to derive data format from file name: {path}")

    @classmethod
    def from_url(cls, url: t.Union[URL, str]) -> "DataFormat":
        if isinstance(url, str):
            url: URL = URL(url)
        if format_ := url.query.get("format"):
            if format_ == "lp":
                return cls.LINE_PROTOCOL_UNCOMPRESSED
            elif format_ == "lp.gz":
                return cls.LINE_PROTOCOL_COMPRESSED
            elif format_ == "csv":
                return cls.ANNOTATED_CSV
            else:
                raise NotImplementedError(f"Invalid data format: {format_}")
        try:
            return cls.from_name(url_fullpath(url))
        except ValueError as ex:
            raise ValueError(f"Unable to derive data format from URL filename or query parameter: {url}") from ex


@dataclasses.dataclass
class OutputFile:
    """
    Manage output file and format for `FileAdapter`.
    """

    path: str
    format: DataFormat  # noqa: A003

    @classmethod
    def from_url(cls, url: t.Union[URL, str]) -> "OutputFile":
        if isinstance(url, str):
            url: URL = URL(url)
        if url.scheme == "file":
            return cls(path=url_or_path(url), format=DataFormat.from_url(url))
        else:
            raise NotImplementedError(f"Unknown file output scheme: {url.scheme}")

    @property
    def stream(self) -> t.IO:
        if self.path == "-":
            return sys.stdout
        else:
            return open(self.path, "w")



import contextlib
import logging
import os
import typing as t
from collections import OrderedDict
from pathlib import Path

import fsspec
import pandas as pd
import polars as pl
from influx_line import InfluxLine

logger = logging.getLogger(__name__)


BytesString = t.Union[bytes, str]
BytesStringList = t.List[BytesString]


def open(path: t.Union[Path, str]):  # noqa: A001
    """
    Access a plethora of resources using `fsspec`.
    """
    path = str(path)
    kwargs: t.Dict[str, t.Any] = {}

    # TODO: Also support authenticated S3.
    if path.startswith("s3"):
        kwargs["anon"] = True

    # TODO: Why isn't compression selected transparently?
    if path.endswith(".gz"):
        kwargs["compression"] = "gzip"
    fs = fsspec.open(path, mode="rb", **kwargs).open()
    return fs


def read_lineprotocol(data: t.IO[t.Any]):
    """
    Read stream of InfluxDB line protocol and decode raw data.

    https://docs.influxdata.com/influxdb/latest/reference/syntax/line-protocol/
    """
    from line_protocol_parser import LineFormatError, parse_line

    for line in data.readlines():
        try:
            yield parse_line(line)
        except LineFormatError as ex:
            logger.info(f"WARNING: Line protocol item {line} invalid. Reason: {ex}")


def records_from_lineprotocol(data: t.IO[t.Any]):
    """
    Read stream of InfluxDB line protocol and generate `OrderedDict` records.
    """
    for lp in read_lineprotocol(data=data):
        record = OrderedDict()
        record["measurement"] = lp["measurement"]
        record["time"] = lp["time"]
        for tag, value in lp["tags"].items():
            record[tag] = value
        for field, value in lp["fields"].items():
            record[field] = value
        yield record


def dataframes_from_lineprotocol(data: t.IO[t.Any]) -> t.Dict[str, pd.DataFrame]:
    """
    Read InfluxDB line protocol file, grouping individual measurement records into multiple Polars DataFrames.
    """
    records = records_from_lineprotocol(data)
    buffer = {}
    frame = pl.DataFrame(records)
    measurements = frame.unique("measurement")["measurement"]
    for measurement in measurements:
        buffer[measurement] = frame.filter(pl.col("measurement") == measurement)
    return buffer


def dataframe_to_lineprotocol(df: pd.DataFrame, progress: bool = False) -> t.Generator[str, None, None]:
    """
    Convert DataFrame to InfluxDB Line Protocol.

    TODO: Needs a test verifying dispatching of tags.
    TODO: Needs configurability to manually dispatch columns to either fields or tags.
    TODO: Needs heuristics if timestamp field is called differently than `time`.
    """
    for record in df.to_dict(orient="records"):
        line = InfluxLine(record["measurement"])
        line.set_timestamp(record["time"].to_datetime64().view("int64"))
        del record["measurement"]
        del record["time"]
        for key, value in record.items():
            if isinstance(value, (int, float)):
                line.add_field(key, value)
            else:
                line.add_tag(key, value)
        yield str(line)


def dataframe_to_sql(
    df: t.Union[pd.DataFrame, pl.DataFrame],
    dburi: str,
    tablename: str,
    schema: str = None,
    index=False,
    chunksize=None,
    if_exists="fail",
    npartitions: int = None,
    progress: bool = False,
):
    """
    Load pandas dataframe into database using Dask.

    https://stackoverflow.com/questions/62404502/using-dasks-new-to-sql-for-improved-efficiency-memory-speed-or-alternative-to

    if_exists : {'fail', 'replace', 'append'}, default 'fail'
        How to behave if the table already exists.

        * fail: Raise a ValueError.
        * replace: Drop the table before inserting new values.
        * append: Insert new values to the existing table.
    """
    logger.info(f"Writing dataframe to SQL: uri={dburi}, table={tablename}")
    import dask.dataframe as dd

    # Set a few defaults.
    if_exists = if_exists or "fail"
    chunksize = chunksize or 5_000
    npartitions = npartitions or int(os.cpu_count() / 2)

    # Optionally enable progress bar.
    ctx = contextlib.nullcontext
    if progress:
        from dask.diagnostics import ProgressBar

        ctx = ProgressBar

    if dburi.startswith("crate"):

        # Use performance INSERT method.
        try:
            from sqlalchemy_cratedb.support import insert_bulk
        except ImportError:  # pragma: nocover
            from crate.client.sqlalchemy.support import insert_bulk

        method = insert_bulk
    else:
        method = "multi"

    # Load data into database.
    with ctx():
        if isinstance(df, pl.DataFrame):
            df = df.to_pandas()
        ddf = dd.from_pandas(df, npartitions=npartitions)
        ddf = ddf.drop(columns="measurement", errors="ignore")
        return ddf.to_sql(
            tablename,
            uri=dburi,
            schema=schema,
            index=index,
            chunksize=chunksize,
            if_exists=if_exists,
            method=method,
            parallel=True,
        )



import sys
import typing

if typing.TYPE_CHECKING:
    from importlib.metadata import EntryPoints


def entry_points(**params) -> "EntryPoints":
    if sys.version_info < (3, 10):
        from importlib_metadata import entry_points
    else:
        from importlib.metadata import entry_points
    return entry_points(**params)



import logging
import textwrap
import typing as t

import click

from influxio.util.common import setup_logging

logger = logging.getLogger(__name__)


def boot_click(ctx: click.Context, verbose: bool = False, debug: bool = False):
    """
    Bootstrap the CLI application.
    """

    # Adjust log level according to `verbose` / `debug` flags.
    log_level = logging.WARNING
    if verbose:
        log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG

    # Setup logging, according to `verbose` / `debug` flags.
    setup_logging(level=log_level)


def split_list(value: str, delimiter: str = ",") -> t.List[str]:
    if value is None:
        return []
    return [c.strip() for c in value.split(delimiter)]


def to_list(x: t.Any, default: t.List[t.Any] = None) -> t.List[t.Any]:
    if not isinstance(default, t.List):
        raise ValueError("Default value is not a list")
    if x is None:
        return default
    if not isinstance(x, t.Iterable) or isinstance(x, str):
        return [x]
    elif isinstance(x, list):
        return x
    else:
        return list(x)


def docstring_format_verbatim(text: t.Optional[str]) -> str:
    """
    Format docstring to be displayed verbatim as a help text by Click.

    - https://click.palletsprojects.com/en/8.1.x/documentation/#preventing-rewrapping
    - https://github.com/pallets/click/issues/56
    """
    text = text or ""
    text = textwrap.dedent(text)
    lines = [line if line.strip() else "\b" for line in text.splitlines()]
    return "\n".join(lines)



import logging
import random
import typing as t
from collections import OrderedDict
from inspect import signature

import numpy as np
import pandas as pd
from pandas.io.formats.info import DataFrameInfo
from pueblo.testing.pandas import makeMixedDataFrame, makeTimeDataFrame

logger = logging.getLogger(__name__)


default_rows_count = 150_000


class DataFrameFactory:
    def __init__(self, rows: int = default_rows_count, columns: int = 99):
        self.rows = int(rows)
        self.columns = int(columns)

    def make(self, which: str, **options) -> pd.DataFrame:
        fun: t.Callable = getattr(self, f"make_{which}")
        sig = signature(fun)

        kwargs = OrderedDict()
        if "rows" in sig.parameters:
            kwargs["rows"] = self.rows
        if "columns" in sig.parameters:
            kwargs["columns"] = self.columns

        logger.info("Creating data frame")
        df = fun(**kwargs)

        # Debug
        df.info()
        info = DataFrameInfo(data=df)
        logger.info(f"Memory usage: {info.memory_usage_string.strip()}")

        return df

    @staticmethod
    def make_dummy() -> pd.DataFrame:
        records = [{"foo": "baz", "bar": "qux"}]
        return pd.DataFrame.from_records(records)

    @staticmethod
    def make_mixed() -> pd.DataFrame:
        return makeMixedDataFrame()

    def make_dateindex(self) -> pd.DataFrame:
        return makeTimeDataFrame(nper=self.rows, freq="s")

    @staticmethod
    def make_wide(rows: int = default_rows_count, columns: int = 99) -> pd.DataFrame:
        """
        https://github.com/influxdata/influxdb-client-python/blob/master/examples/ingest_large_dataframe.py
        """
        col_data = {
            "time": np.arange(0, rows, 1, dtype=int),
            "tag": np.random.choice(["tag_a", "tag_b", "test_c"], size=(rows,)),
        }
        for n in range(2, columns + 1):
            col_data[f"col{n}"] = random.randint(1, 10)  # noqa: S311

        data_frame = pd.DataFrame(data=col_data).set_index("time")
        return data_frame



# This is a shim to allow GitHub to decode the package in order to provide
# the "Used by" section on the project homepage.
import setuptools

if __name__ == "__main__":
    setuptools.setup(name="influxio")
